import streamlit as st
import google.generativeai as genai 
from qdrant_client import QdrantClient, models
import pdfplumber
import datetime
import pandas as pd
import uuid
import os
from openai import OpenAI
from PIL import Image
import base64
from io import BytesIO
from dotenv import load_dotenv
load_dotenv()

# API 키 설정
client = OpenAI(
    api_key=os.getenv("OPENAI_API_KEY"),
    base_url="https://api.openai.com/v1"
)

# Qdrant 클라이언트 설정
qdrant_client = QdrantClient(
    url=os.getenv("QDRANT_URL"), 
    api_key=os.getenv("QDRANT_API_KEY")
)

# Gemini 설정
genai.configure(api_key=os.getenv("GEMINI_API_KEY"))
model = genai.GenerativeModel('gemini-2.0-flash-exp')

class Notebook:
    def __init__(self, name):
        self.name = name
        self.contents = []
        
    def add_content(self, content_type, content, metadata=None):
        if content_type == 'pdf':
            self.contents.append({
                "type": content_type,
                "content": content,
                "metadata": metadata
            })
        else:
            self.contents.append({
                "type": content_type,
                "content": content,
                "metadata": metadata
            })

def init_collections():
    collections_config = {
        "notebooks": 1536,
        "statements": 1536,
        "schedules": 1536, 
        "documents": 1536,
        "chats": 1536
    }
    
    try:
        collections = qdrant_client.get_collections().collections
        collection_names = [c.name for c in collections]
        
        for name, size in collections_config.items():
            if name not in collection_names:
                qdrant_client.create_collection(
                    collection_name=name,
                    vectors_config=models.VectorParams(size=size, distance=models.Distance.COSINE)
                )
    except Exception as e:
        st.warning("컬렉션이 이미 존재합니다.")

def text_to_vector(text):
    if text is None or len(text.strip()) == 0:
        text = "empty"
    text = text.replace('\n', ' ').strip()

    response = client.embeddings.create(
        input=text[:8000],
        model="text-embedding-ada-002",
        encoding_format="float"
    )
    return response.data[0].embedding

def save_notebook(notebook):
    if not notebook.contents:
        return None

    existing_notebook = get_notebook(notebook.name)
    
    if existing_notebook:
        vector = text_to_vector(notebook.name)
        point_id = existing_notebook.id
        payload = {
            "notebook_name": notebook.name,
            "contents": notebook.contents.copy(),
            "date": str(datetime.datetime.now())
        }
        point = models.PointStruct(
            id=point_id,
            vector=vector,
            payload=payload
        )
        qdrant_client.upsert(
            collection_name="notebooks",
            points=[point]
        )
        return point_id
    else:
        vector = text_to_vector(notebook.name)
        point_id = str(uuid.uuid4())

        payload = {
            "notebook_name": notebook.name,
            "contents": notebook.contents.copy(),
            "date": str(datetime.datetime.now())
        }

        point = models.PointStruct(
            id=point_id,
            vector=vector,
            payload=payload
        )

        qdrant_client.upsert(
            collection_name="notebooks",
            points=[point]
        )

        return point_id

def get_notebook(notebook_name):
    results = qdrant_client.scroll(
        collection_name="notebooks",
        scroll_filter=models.Filter(
            must=[
                models.FieldCondition(
                    key="notebook_name",
                    match=models.MatchValue(value=notebook_name)
                )
            ]
        )
    )[0]
    
    return results[0] if results else None

def extract_pdf_content(uploaded_file):
    text_content = []
    
    with pdfplumber.open(uploaded_file) as pdf:
        for page in pdf.pages:
            # 텍스트 추출
            text = page.extract_text()
            if text:
                text_content.append(text)
            
            # 표 데이터 추출
            tables = page.extract_tables()
            for table in tables:
                # None 값 처리
                processed_table = [[str(cell) if cell is not None else '' for cell in row] for row in table]
                table_text = "\n".join([" | ".join(row) for row in processed_table if any(row)])
                if table_text:
                    text_content.append(f"표 데이터:\n{table_text}")
    
    # 전체 텍스트 결합
    full_content = "\n\n".join(text_content)
    return full_content, uploaded_file.name # 두 번째 반환값은 파일명

def chat_with_notebook(notebook_name, user_input):
    notebook = get_notebook(notebook_name)
    if not notebook:
        return "노트북을 찾을 수 없습니다."

    model = genai.GenerativeModel('gemini-2.0-flash-exp')
    chat = model.start_chat(history=[])
    
    # 컨텐츠 처리
    message_parts = []
    content_summary = []
    
    # 기본 정보
    content_summary.append(f"=== 노트북 정보 ===")
    content_summary.append(f"제목: {notebook.payload['notebook_name']}")
    content_summary.append(f"생성일: {notebook.payload['date']}")
    content_summary.append(f"\n=== 노트북 내용 ===")
    
    # 컨텐츠 카운터
    text_count = 0
    pdf_count = 0
    image_count = 0
    
    # 각 컨텐츠 처리
    for content in notebook.payload["contents"]:
        if content["type"] == "text":
            text_count += 1
            content_summary.append(f"\n[텍스트 문서 {text_count}]")
            content_summary.append(content["content"])
            
        elif content["type"] == "pdf":
            pdf_count += 1
            content_summary.append(f"\n[PDF 문서 {pdf_count}]")
            content_summary.append(content["content"])
            
        elif content["type"] == "image":
            image_count += 1
            try:
                image_bytes = base64.b64decode(content["content"])
                image = Image.open(BytesIO(image_bytes))
                message_parts.append({
                    "mime_type": "image/png",
                    "data": base64.b64encode(image_bytes).decode()
                })
                content_summary.append(f"\n[이미지 {image_count}] 파일명: {content['metadata']}")
            except Exception as e:
                print(f"이미지 처리 오류: {str(e)}")

    # 컨텐츠 요약 추가
    summary = []
    if text_count > 0:
        summary.append(f"텍스트 문서 {text_count}개")
    if pdf_count > 0:
        summary.append(f"PDF 문서 {pdf_count}개")
    if image_count > 0:
        summary.append(f"이미지 {image_count}개")
    
    if summary:
        content_summary.insert(4, "포함된 컨텐츠: " + ", ".join(summary))

    # 메시지 구성
    prompt = "\n".join(content_summary)
    prompt += f"\n\n질문: {user_input}\n"
    prompt += "위의 노트북 내용을 모두 참고하여 자세히 답변해주세요."
    
    message_parts.insert(0, {"text": prompt})
    
    # 응답 생성
    response = chat.send_message(message_parts, stream=True)
    return response

def delete_notebook(notebook_name):
    try:
        # 노트북 검색
        results = qdrant_client.scroll(
            collection_name="notebooks",
            scroll_filter=models.Filter(
                must=[
                    models.FieldCondition(
                        key="notebook_name",
                        match=models.MatchValue(value=notebook_name)
                    )
                ]
            )
        )[0]
        
        # 노트북 삭제
        if results:
            point_id = results[0].id
            qdrant_client.delete(
                collection_name="notebooks", 
                points_selector=models.PointIdsList(points=[point_id])
            )
            return True
        return False
        
    except Exception as e:
        print(f"삭제 중 오류 발생: {e}")
        return False

def get_all_notebooks():
    notebooks = qdrant_client.scroll(
        collection_name="notebooks",
        limit=100  # 충분한 수의 노트북을 가져오도록 설정
    )[0]
    return notebooks

def save_chat(notebook_name, chat_id, messages):
    vector = text_to_vector(notebook_name + chat_id)
    payload = {
        "notebook_name": notebook_name,
        "chat_id": chat_id,
        "messages": messages,
        "date": str(datetime.datetime.now())
    }
    
    point = models.PointStruct(
        id=chat_id,
        vector=vector,
        payload=payload
    )
    
    qdrant_client.upsert(
        collection_name="chats",
        points=[point]
    )

def get_chats(notebook_name):
    results = qdrant_client.scroll(
        collection_name="chats",
        scroll_filter=models.Filter(
            must=[
                models.FieldCondition(
                    key="notebook_name",
                    match=models.MatchValue(value=notebook_name)
                )
            ]
        )
    )[0]
    return results

def delete_chat(chat_id):
    qdrant_client.delete(
        collection_name="chats",
        points_selector=models.PointIdsList(points=[chat_id])
    )



def main():
    # main() 함수 시작 부분에 추가
    if 'notebook_chats' not in st.session_state:
        st.session_state.notebook_chats = {}  # 노트북별 채팅 기록 저장
    if 'current_chat' not in st.session_state:
        st.session_state.current_chat = None  # 현재 선택된 채팅

    if 'chat_history' not in st.session_state:
        st.session_state.chat_history = []
        
    st.title("Gemini 2.0 Flash Exp 멀티모달 챗봇")
    
    menu = st.sidebar.selectbox(
        "메뉴 선택",
        ["노트북 관리", "노트북 수정", "노트북 채팅"]
    )
    
    if menu == "노트북 관리":
        st.header("노트북 관리")
        
        # 탭 생성 코드 추가
        tab1, tab2, tab3 = st.tabs(["노트북 목록", "노트북 생성", "노트북 삭제"])
        
        with tab1:
            # 실시간으로 노트북 목록 새로고침
            notebooks = get_all_notebooks()
            st.write(f"총 {len(notebooks)}개의 노트북이 있습니다.")
            
            for notebook in notebooks:
                with st.expander(f"📚 {notebook.payload['notebook_name']}"):
                    st.write(f"생성일: {notebook.payload['date']}")
                    for content in notebook.payload["contents"]:
                        if content['type'] == 'text':
                            st.write(f"텍스트: {content['content'][:200]}...")
                        elif content['type'] == 'image':
                            st.write(f"이미지: {content['metadata']}")
                            # 이미지 미리보기 추가
                            image_bytes = base64.b64decode(content['content'])
                            st.image(image_bytes, caption=content['metadata'])
                        elif content['type'] == 'pdf':
                            st.write(f"PDF: {content['metadata']}")
                            st.write(f"미리보기: {content['content'][:200]}...")
        with tab2:
            st.subheader("새 노트북 생성")
            
            # 세션 상태 초기화
            if 'current_notebook' not in st.session_state:
                st.session_state.current_notebook = None
                
            notebook_name = st.text_input("노트북 이름", placeholder="제목을 입력하면 텍스트, PDF, 이미지를 추가할수 있습니다.")
            
            if notebook_name:
                if st.session_state.current_notebook is None or st.session_state.current_notebook.name != notebook_name:
                    st.session_state.current_notebook = Notebook(notebook_name)
                    
                col1, col2, col3 = st.columns(3)
                
                with col1:
                    text_content = st.text_area("텍스트 내용")
                    if st.button("텍스트 추가", key="add_text"):
                        st.session_state.current_notebook.add_content("text", text_content)
                        st.success(f"텍스트가 추가되었습니다. (총 {len(st.session_state.current_notebook.contents)}개 컨텐츠)")
                
                with col2:
                    pdf_file = st.file_uploader("PDF 파일", type="pdf", key="pdf_upload")
                    if pdf_file and st.button("PDF 추가", key="add_pdf"):
                        pdf_content, pdf_filename = extract_pdf_content(pdf_file)
                        st.session_state.current_notebook.add_content("pdf", pdf_content, metadata=pdf_filename)
                        st.success(f"PDF가 추가되었습니다. (총 {len(st.session_state.current_notebook.contents)}개 컨텐츠)")
                
                with col3:
                    image_file = st.file_uploader("이미지", type=["jpg", "png"], key="image_upload")
                    if image_file and st.button("이미지 추가", key="add_image"):
                        image = Image.open(image_file)
                        img_byte_arr = BytesIO()
                        image.save(img_byte_arr, format='PNG')
                        img_base64 = base64.b64encode(img_byte_arr.getvalue()).decode()
                        st.session_state.current_notebook.add_content("image", img_base64, metadata=image_file.name)
                        st.success(f"이미지가 추가되었습니다. (총 {len(st.session_state.current_notebook.contents)}개 컨텐츠)")

                # 현재 추가된 컨텐츠 표시
                if st.session_state.current_notebook and st.session_state.current_notebook.contents:
                    st.write("---")
                    st.write("📚 현재 추가된 컨텐츠:")
                    for idx, content in enumerate(st.session_state.current_notebook.contents, 1):
                        st.write(f"{idx}. {content['type']}")
                    st.write("---")

                if st.button("노트북 저장", type="primary", key="save_notebook"):
                    if st.session_state.current_notebook and st.session_state.current_notebook.contents:
                        point_id = save_notebook(st.session_state.current_notebook)
                        if point_id:
                            st.success(f"노트북이 성공적으로 저장되었습니다! (ID: {point_id})")
                            st.session_state.current_notebook = None
                            st.rerun()
                    else:
                        st.warning("노트북에 내용을 추가한 후 저장해주세요.")
        
        with tab3:
            st.subheader("노트북 삭제")
            notebooks = qdrant_client.scroll(collection_name="notebooks")[0]
            notebook_names = [n.payload["notebook_name"] for n in notebooks]
            
            if notebook_names:
                notebook_to_delete = st.selectbox("삭제할 노트북 선택", notebook_names)
                if st.button("노트북 삭제", type="primary"):
                    if delete_notebook(notebook_to_delete):
                        st.success(f"{notebook_to_delete} 노트북이 삭제되었습니다.")
                        st.rerun()
                    else:
                        st.error("노트북 삭제 중 오류가 발생했습니다.")
            else:
                st.info("삭제할 노트북이 없습니다.")
    
    elif menu == "노트북 수정":
        st.header("노트북 수정")
        notebooks = get_all_notebooks()
        notebook_names = [n.payload["notebook_name"] for n in notebooks]
        
        if notebook_names:
            selected_notebook = st.selectbox("수정할 노트북 선택", notebook_names)
            
            if selected_notebook:
                notebook_record = get_notebook(selected_notebook)
                if notebook_record:
                    notebook = Notebook(notebook_record.payload["notebook_name"])
                    notebook.contents = notebook_record.payload["contents"]
                    
                    st.write(f"선택된 노트북: {notebook.name}")
                    
                    # 컨텐츠 표시를 expander로 변경
                    with st.expander("현재 컨텐츠 목록", expanded=True):
                        for idx, content in enumerate(notebook.contents):
                            if content['type'] == 'image':
                                st.write(f"컨텐츠 {idx+1}: 이미지 - {content['metadata']}")
                                image_bytes = base64.b64decode(content['content'])
                                st.image(image_bytes)
                            elif content['type'] == 'pdf':
                                st.write(f"컨텐츠 {idx+1}: PDF - {content['metadata']}")
                                st.write(content['content'][:200] + "..." if len(content['content']) > 200 else content['content'])
                            else:
                                st.write(f"컨텐츠 {idx+1}: 텍스트")
                                st.write(content['content'][:200] + "..." if len(content['content']) > 200 else content['content'])
                    
                    # 컨텐츠 추가/삭제 기능
                    tab1, tab2 = st.tabs(["컨텐츠 추가", "컨텐츠 삭제"])
                    
                    with tab1:
                        col1, col2, col3 = st.columns(3)
                        
                        with col1:
                            text_content = st.text_area("텍스트 내용", key=f"edit_text_{selected_notebook}")
                            if st.button("텍스트 추가", key=f"add_text_{selected_notebook}"):
                                notebook.add_content("text", text_content)
                                save_notebook(notebook)
                                st.success("텍스트가 추가되었습니다.")
                                st.rerun()
                        
                        with col2:
                            pdf_file = st.file_uploader("PDF 파일", type="pdf", key=f"edit_pdf_{selected_notebook}")
                            if pdf_file and st.button("PDF 추가", key=f"add_pdf_{selected_notebook}"):
                                pdf_content, pdf_filename = extract_pdf_content(pdf_file)
                                notebook.add_content("pdf", pdf_content, metadata=pdf_filename)
                                save_notebook(notebook)
                                st.success("PDF가 추가되었습니다.")
                                st.rerun()
                        
                        with col3:
                            image_file = st.file_uploader("이미지", type=["jpg", "png"], key=f"edit_image_{selected_notebook}")
                            if image_file and st.button("이미지 추가", key=f"add_image_{selected_notebook}"):
                                image = Image.open(image_file)
                                img_byte_arr = BytesIO()
                                image.save(img_byte_arr, format='PNG')
                                img_base64 = base64.b64encode(img_byte_arr.getvalue()).decode()
                                notebook.add_content("image", img_base64, metadata=image_file.name)
                                save_notebook(notebook)
                                st.success("이미지가 추가되었습니다.")
                                st.rerun()
                    
                    with tab2:
                        content_labels = []
                        for i, c in enumerate(notebook.contents):
                            if c['type'] == 'text':
                                label = f"컨텐츠 {i+1} (텍스트)"
                            elif c['type'] == 'pdf':
                                label = f"컨텐츠 {i+1} (PDF: {c['metadata']})"
                            elif c['type'] == 'image':
                                label = f"컨텐츠 {i+1} (이미지: {c['metadata']})"
                            content_labels.append(label)

                        if content_labels:
                            content_to_delete = st.selectbox("삭제할 컨텐츠 선택", content_labels, key=f"delete_content_{selected_notebook}")
                            if st.button("컨텐츠 삭제", type="primary", key=f"delete_button_{selected_notebook}"):
                                index_to_delete = int(content_to_delete.split(" ")[1]) - 1
                                del notebook.contents[index_to_delete]
                                save_notebook(notebook)
                                st.success("컨텐츠가 삭제되었습니다.")
                                st.rerun()
                        else:
                            st.info("삭제할 컨텐츠가 없습니다.")
                else:
                    st.error("노트북을 찾을 수 없습니다.")
        else:
            st.info("수정할 노트북이 없습니다.")

    elif menu == "노트북 채팅":
        st.header("노트북 기반 챗봇")
        
        notebooks = qdrant_client.scroll(collection_name="notebooks")[0]
        notebook_names = [n.payload["notebook_name"] for n in notebooks]
        
        if notebooks:
            with st.sidebar:
                selected_notebook = st.selectbox("노트북 선택", notebook_names)
                
                if st.button("새 대화 시작"):
                    chat_id = str(uuid.uuid4())
                    if selected_notebook not in st.session_state.notebook_chats:
                        st.session_state.notebook_chats[selected_notebook] = {}
                    st.session_state.notebook_chats[selected_notebook][chat_id] = []
                    st.session_state.current_chat = (selected_notebook, chat_id)
                    save_chat(selected_notebook, chat_id, [])
                    st.rerun()
                
                st.markdown("---")
                st.write("이전 대화:")
                
                # 저장된 채팅 불러오기
                chats = get_chats(selected_notebook)
                for chat in chats:
                    preview = chat.payload['messages'][0]['content'][:30] + "..." if chat.payload['messages'] else "새 대화"
                    col1, col2 = st.columns([4, 1])
                    with col1:
                        if st.button(f"💬 {preview}", key=f"chat_{chat.payload['chat_id']}"):
                            st.session_state.current_chat = (selected_notebook, chat.payload['chat_id'])
                            st.session_state.notebook_chats[selected_notebook] = {
                                chat.payload['chat_id']: chat.payload['messages']
                            }
                            st.rerun()
                    with col2:
                        if st.button("🗑️", key=f"delete_{chat.payload['chat_id']}"):
                            delete_chat(chat.payload['chat_id'])
                            if st.session_state.current_chat and st.session_state.current_chat[1] == chat.payload['chat_id']:
                                st.session_state.current_chat = None
                            st.rerun()
            
            # 메인 채팅 인터페이스
            if st.session_state.current_chat and st.session_state.current_chat[0] == selected_notebook:
                notebook, chat_id = st.session_state.current_chat
                messages = st.session_state.notebook_chats[notebook][chat_id]
                
                # 채팅 표시
                chat_container = st.container()
                with chat_container:
                    for msg in messages:
                        role = "사용자" if msg['role'] == "user" else "AI"
                        st.write(f"{role}: {msg['content']}")
                
                # 입력 영역
                def handle_input():
                    if user_input := st.session_state.user_input:
                        messages.append({"role": "user", "content": user_input})
                        save_chat(notebook, chat_id, messages)
                        st.session_state.user_input = ""
                        st.session_state.need_response = True
                
                st.text_input("질문 입력", key="user_input", on_change=handle_input)
                
                # 새로운 응답 생성
                if 'need_response' in st.session_state and st.session_state.need_response:
                    last_user_message = messages[-1]['content']
                    response_container = st.empty()
                    full_response = ""
                    
                    for chunk in chat_with_notebook(selected_notebook, last_user_message):
                        full_response += chunk.text
                        response_container.markdown(full_response + "▌")
                    
                    messages.append({"role": "assistant", "content": full_response})
                    save_chat(notebook, chat_id, messages)
                    st.session_state.need_response = False
                    st.rerun()
            else:
                st.info("사이드바에서 새 대화를 시작하거나 이전 대화를 선택해주세요.")
        else:
            st.info("저장된 노트북이 없습니다. 먼저 노트북을 생성해주세요.")


if __name__ == "__main__":
    init_collections()
    main()
